from . import BASIC, COMFORT, NS_RAM, NS_QDT
from .elements import Element
from .fields import BinaryObjectField, DateTimeField, StringField, IDField


class ProcuringProjectType(Element):
    id = StringField(NS_RAM, "ID")
    name = StringField(NS_RAM, "Name")

    class Meta:
        namespace = NS_RAM
        tag = "SpecifiedProcuringProject"


class ReferencedDocument(Element):
    issuer_assigned_id = IDField(
        NS_RAM, "IssuerAssignedID", required=False, profile=BASIC
    )
    issue_date_time = DateTimeField(
        NS_RAM,
        "FormattedIssueDateTime",
        required=True,
        profile=BASIC,
        date_time_namespace=NS_QDT,
    )


class SellerOrderReferencedDocument(ReferencedDocument):
    class Meta:
        namespace = NS_RAM
        tag = "SellerOrderReferencedDocument"


class BuyerOrderReferencedDocument(ReferencedDocument):
    class Meta:
        namespace = NS_RAM
        tag = "BuyerOrderReferencedDocument"


class ContractReferencedDocument(ReferencedDocument):
    class Meta:
        namespace = NS_RAM
        tag = "ContractReferencedDocument"


class AdditionalReferencedDocument(ReferencedDocument):
    uri_id = StringField(NS_RAM, "URIID", required=False, profile=COMFORT)
    type_code = StringField(NS_RAM, "TypeCode", profile=COMFORT, required=True)
    name = StringField(NS_RAM, "Name", profile=COMFORT, required=False)
    attached_object = BinaryObjectField(
        NS_RAM, "AttachmentBinaryObject", required=False, profile=COMFORT
    )

    class Meta:
        namespace = NS_RAM
        tag = "AdditionalReferencedDocument"


class InvoiceReferencedDocument(ReferencedDocument):
    type_code = StringField(NS_RAM, "TypeCode", profile=COMFORT, required=False)

    class Meta:
        namespace = NS_RAM
        tag = "InvoiceReferencedDocument"


class InvoiceSpecifiedReferencedDocument(ReferencedDocument):
    type_code = StringField(NS_RAM, "TypeCode", profile=COMFORT, required=False)

    class Meta:
        namespace = NS_RAM
        tag = "InvoiceSpecifiedReferencedDocument"


class UltimateCustomerOrderReferencedDocument(ReferencedDocument):
    class Meta:
        namespace = NS_RAM
        tag = "UltimateCustomerOrderReferencedDocument"


class DespatchAdviceReferencedDocument(ReferencedDocument):
    class Meta:
        namespace = NS_RAM
        tag = "DespatchAdviceReferencedDocument"


class LineUltimateCustomerOrderReferencedDocument(ReferencedDocument):
    line_id = StringField(NS_RAM, "LineID", required=False, profile=COMFORT)

    class Meta:
        namespace = NS_RAM
        tag = "UltimateCustomerOrderReferencedDocument"


class LineBuyerOrderReferencedDocument(ReferencedDocument):
    line_id = StringField(NS_RAM, "LineID", required=False, profile=COMFORT)

    class Meta:
        namespace = NS_RAM
        tag = "BuyerOrderReferencedDocument"


class LineContractReferencedDocument(ReferencedDocument):
    line_id = StringField(NS_RAM, "LineID", required=False, profile=COMFORT)

    class Meta:
        namespace = NS_RAM
        tag = "ContractReferencedDocument"


class LineDespatchAdviceReferencedDocument(ReferencedDocument):
    line_id = StringField(NS_RAM, "LineID", required=False, profile=COMFORT)

    class Meta:
        namespace = NS_RAM
        tag = "DespatchAdviceReferencedDocument"


class ReceivingAdviceReferencedDocument(ReferencedDocument):
    class Meta:
        namespace = NS_RAM
        tag = "ReceivingAdviceReferencedDocument"


class LineReceivingAdviceReferencedDocument(ReferencedDocument):
    line_id = StringField(NS_RAM, "LineID", required=False, profile=COMFORT)

    class Meta:
        namespace = NS_RAM
        tag = "ReceivingAdviceReferencedDocument"


class LineAdditionalReferencedDocument(ReferencedDocument):
    uri_id = StringField(NS_RAM, "URIID", required=False, profile=COMFORT)
    line_id = StringField(NS_RAM, "LineID", required=False, profile=COMFORT)
    type_code = StringField(NS_RAM, "TypeCode", required=False, profile=COMFORT)
    name = StringField(NS_RAM, "Name", required=False, profile=COMFORT)
    reference_type_code = StringField(
        NS_RAM, "ReferenceTypeCode", profile=COMFORT, required=True
    )
    attached_object = BinaryObjectField(
        NS_RAM, "AttachmentBinaryObject", required=False, profile=COMFORT
    )

    class Meta:
        namespace = NS_RAM
        tag = "AdditionalReferencedDocument"


class DeliveryNoteReferencedDocument(ReferencedDocument):
    class Meta:
        namespace = NS_RAM
        tag = "DeliveryNoteReferencedDocument"


class LineDeliveryNoteReferencedDocument(ReferencedDocument):
    class Meta:
        namespace = NS_RAM
        tag = "DeliveryNoteReferencedDocument"



from . import BASIC, COMFORT, EXTENDED, NS_RAM
from .accounting import (
    ApplicableTradeTax,
    BillingSpecifiedPeriod,
    ReceivableAccountingAccount,
    TradeAllowanceCharge,
)
from .container import Container
from .delivery import SupplyChainEvent
from .elements import Element
from .fields import (
    DecimalField,
    Field,
    MultiField,
    QuantityField,
    StringField,
)
from .note import IncludedNote
from .party import ShipToTradeParty, UltimateShipToTradeParty
from .product import TradeProduct
from .references import (
    InvoiceReferencedDocument,
    LineAdditionalReferencedDocument,
    LineBuyerOrderReferencedDocument,
    LineContractReferencedDocument,
    LineDeliveryNoteReferencedDocument,
    LineDespatchAdviceReferencedDocument,
    LineReceivingAdviceReferencedDocument,
    LineUltimateCustomerOrderReferencedDocument,
)


class AllowanceCharge(TradeAllowanceCharge):
    class Meta:
        namespace = NS_RAM
        tag = "AppliedTradeAllowanceCharge"


class GrossPrice(Element):
    amount = DecimalField(
        NS_RAM, "ChargeAmount", required=True, profile=COMFORT, _d="Bruttopreis"
    )
    basis_quantity = QuantityField(
        NS_RAM, "BasisQuantity", required=False, profile=COMFORT, _d="Preisbasismenge"
    )
    charge: Container = MultiField(AllowanceCharge, required=False, profile=COMFORT)

    class Meta:
        namespace = NS_RAM
        tag = "GrossPriceProductTradePrice"


class NetPrice(Element):
    amount = DecimalField(NS_RAM, "ChargeAmount", required=True, profile=COMFORT)
    basis_quantity = QuantityField(
        NS_RAM, "BasisQuantity", required=False, profile=COMFORT, _d="Preisbasismenge"
    )
    # TODO: IncludedTradeTax missing

    class Meta:
        namespace = NS_RAM
        tag = "NetPriceProductTradePrice"


class LineDocument(Element):
    line_id = StringField(NS_RAM, "LineID")
    line_status_code = StringField(NS_RAM, "LineStatusCode")
    line_status_reason_code = StringField(NS_RAM, "LineStatusReasonCode")
    notes: Container = MultiField(IncludedNote)

    class Meta:
        namespace = NS_RAM
        tag = "AssociatedDocumentLineDocument"


class LineAgreement(Element):
    buyer_order: LineBuyerOrderReferencedDocument = Field(
        LineBuyerOrderReferencedDocument, required=False, profile=EXTENDED
    )
    contract: LineContractReferencedDocument = Field(
        LineContractReferencedDocument, required=False, profile=EXTENDED
    )
    customer_order: LineUltimateCustomerOrderReferencedDocument = Field(
        LineUltimateCustomerOrderReferencedDocument, required=False, profile=EXTENDED
    )
    additional_references: Container = MultiField(
        LineAdditionalReferencedDocument, required=False, profile=COMFORT
    )
    gross: GrossPrice = Field(GrossPrice, required=False, profile=COMFORT)
    net: NetPrice = Field(NetPrice)

    class Meta:
        namespace = NS_RAM
        tag = "SpecifiedLineTradeAgreement"


class LineDelivery(Element):
    billed_quantity = QuantityField(
        NS_RAM, "BilledQuantity", required=True, profile=BASIC, _d="Menge, berechnet"
    )
    charge_free_quantity = QuantityField(
        NS_RAM,
        "ChargeFreeQuantity",
        required=False,
        profile=EXTENDED,
        _d="Menge, ohne Berechnung",
    )
    package_quantity = QuantityField(
        NS_RAM,
        "PackageQuantity",
        required=False,
        profile=EXTENDED,
        _d="Anzahl Packstücke",
    )
    ship_to: ShipToTradeParty = Field(
        ShipToTradeParty, required=False, profile=EXTENDED
    )
    ultimate_ship_to: UltimateShipToTradeParty = Field(
        UltimateShipToTradeParty, required=False, profile=EXTENDED
    )
    event: SupplyChainEvent = Field(
        SupplyChainEvent,
        required=False,
        profile=EXTENDED,
        _d="Detailinformationen zur tatsächlichen Lieferung",
    )
    despatch_advice: LineDespatchAdviceReferencedDocument = Field(
        LineDespatchAdviceReferencedDocument, required=False, profile=EXTENDED
    )
    receiving_advice: LineReceivingAdviceReferencedDocument = Field(
        LineReceivingAdviceReferencedDocument, required=False, profile=EXTENDED
    )
    delivery_note: LineDeliveryNoteReferencedDocument = Field(
        LineDeliveryNoteReferencedDocument, required=False, profile=EXTENDED
    )

    class Meta:
        namespace = NS_RAM
        tag = "SpecifiedLineTradeDelivery"


class LineSummation(Element):
    total_amount = DecimalField(
        NS_RAM, "LineTotalAmount", required=True, profile=COMFORT
    )
    total_allowance_charge = DecimalField(
        NS_RAM,
        "TotalAllowanceChargeAmount",
        required=False,
        profile=EXTENDED,
        _d="Gesamtbetrag der Zu- und Abschläge",
    )

    class Meta:
        namespace = NS_RAM
        tag = "SpecifiedTradeSettlementLineMonetarySummation"


class LineSettlement(Element):
    trade_tax: ApplicableTradeTax = Field(ApplicableTradeTax, required=False)
    period: BillingSpecifiedPeriod = Field(BillingSpecifiedPeriod, required=False)
    allowance_charge: Container = MultiField(
        TradeAllowanceCharge,
        required=False,
        _d="Schalter für Zu-/Abschlag",
    )
    monetary_summation: LineSummation = Field(
        LineSummation, required=False, profile=BASIC
    )
    invoice_referenced_document: Container = MultiField(
        InvoiceReferencedDocument,
        required=False,
        profile=EXTENDED,
        _d="Referenz auf die vorausgegangene Rechnung",
    )
    additional_referenced_document: LineAdditionalReferencedDocument = Field(
        LineAdditionalReferencedDocument, required=False, profile=COMFORT
    )
    accounting_account: ReceivableAccountingAccount = Field(
        ReceivableAccountingAccount,
        required=False,
        profile=COMFORT,
        _d="Detailinformationen zur Buchungsreferenz",
    )

    class Meta:
        namespace = NS_RAM
        tag = "SpecifiedLineTradeSettlement"


class LineItem(Element):
    document: LineDocument = Field(LineDocument, required=True)
    product: TradeProduct = Field(TradeProduct)
    agreement: LineAgreement = Field(LineAgreement)
    delivery: LineDelivery = Field(LineDelivery)
    settlement: LineSettlement = Field(LineSettlement, required=True)

    class Meta:
        namespace = NS_RAM
        tag = "IncludedSupplyChainTradeLineItem"



import pytest

from drafthorse.pdf import attach_xml


@pytest.mark.parametrize("invoice_document", ["380"], indirect=True)
def test_readme_construction_example_pdf16(invoice_document, empty_pdf16_bytes):
    """
    Test using a PDF 1.6 version
    """
    xml = invoice_document.serialize(schema="FACTUR-X_EXTENDED")
    output_pdf = attach_xml(empty_pdf16_bytes, xml)

    assert output_pdf


@pytest.mark.parametrize("invoice_document", ["380"], indirect=True)
def test_readme_construction_example_pdf17(invoice_document, invoice_pdf17_bytes):
    """
    Test using a PDF 1.7 version in order to cover the output intents handling
    """
    xml = invoice_document.serialize(schema="FACTUR-X_EXTENDED")
    output_pdf = attach_xml(invoice_pdf17_bytes, xml)

    assert output_pdf


@pytest.mark.parametrize("invoice_document", ["380"], indirect=True)
def test_readme_construction_example_pdf17_en16931(
    invoice_document, invoice_pdf17_bytes
):
    """
    Test using a PDF 1.7 version in order to cover the output intents handling
    """
    doc = invoice_document
    doc.context.guideline_parameter.id = "urn:cen.eu:en16931:2017"
    xml = doc.serialize(schema="FACTUR-X_EXTENDED")
    output_pdf = attach_xml(invoice_pdf17_bytes, xml)

    assert output_pdf


@pytest.mark.parametrize("invoice_document", ["380"], indirect=True)
def test_readme_construction_example_pdf17_xrechnung2(
    invoice_document, invoice_pdf17_bytes
):
    """
    Test using a PDF 1.7 version in order to cover the output intents handling
    """
    doc = invoice_document
    doc.context.guideline_parameter.id = (
        "urn:cen.eu:en16931:2017#compliant#urn:xoev-de:kosit:standard:xrechnung_2.3"
    )
    xml = doc.serialize(schema="FACTUR-X_EXTENDED")
    output_pdf = attach_xml(invoice_pdf17_bytes, xml)

    assert output_pdf


@pytest.mark.parametrize("invoice_document", ["380"], indirect=True)
def test_readme_construction_example_pdf17_xrechnung3(
    invoice_document, invoice_pdf17_bytes
):
    """
    Test using a PDF 1.7 version in order to cover the output intents handling
    """
    doc = invoice_document
    doc.context.guideline_parameter.id = (
        "urn:cen.eu:en16931:2017#compliant#urn:xeinkauf.de:kosit:xrechnung_3.0"
    )
    xml = doc.serialize(schema="FACTUR-X_EXTENDED")
    output_pdf = attach_xml(invoice_pdf17_bytes, xml)

    assert output_pdf


@pytest.mark.parametrize("invoice_document", ["380"], indirect=True)
def test_invalid_invoice_XML_profile_exceptions(invoice_document, invoice_pdf17_bytes):
    """
    Test using a PDF 1.7 version in order to cover the output intents handling
    """
    doc = invoice_document
    doc.context.guideline_parameter.id = (
        "urn:cen.eu:en16932:2017#conformant#urn:factur-x.eu:1p0:wrong"
    )
    xml = doc.serialize(schema="FACTUR-X_EXTENDED")
    # invalid pdf type
    with pytest.raises(Exception) as exc_info:
        attach_xml(invoice_pdf17_bytes, xml)

    assert str(exc_info.value) == "Invalid XML profile!"


@pytest.mark.parametrize("invoice_document", ["220"], indirect=True)
def test_invalid_invoice_exceptions(invoice_document, invoice_pdf17_bytes):
    """
    Test invalid cases
    """
    xml = invoice_document.serialize(schema="FACTUR-X_EXTENDED")

    # invalid pdf type
    with pytest.raises(Exception) as exc_info:
        attach_xml("invalid_pdf_type", xml)

    assert str(exc_info.value) == "Please supply original PDF as bytes."

    # invalid xml type
    with pytest.raises(Exception) as exc_info:
        attach_xml(invoice_pdf17_bytes, "invalid_xml_type")

    assert str(exc_info.value) == "Please supply XML data as bytes."


@pytest.mark.parametrize("invoice_document", ["380"], indirect=True)
def test_unused_name_spaces_included(invoice_document):
    doc = invoice_document
    doc.context.guideline_parameter.id = (
        "urn:cen.eu:en16932:2017#conformant#urn:factur-x.eu:1p0:wrong"
    )
    xml = doc.serialize(schema="FACTUR-X_EXTENDED")
    assert (
        'xmlns:qdt="urn:un:unece:uncefact:data:standard:QualifiedDataType:100"'
        in xml.decode()
    )
    assert 'xmlns:xs="http://www.w3.org/2001/XMLSchema"' in xml.decode()
    assert 'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"' in xml.decode()
    assert "xs:" not in xml.decode()
    assert "xsi:" not in xml.decode()



"""
FACTUR-X XMP with the required PDF/A extension schema description
"""

XMP_SCHEMA = """
<?xpacket begin="\ufeff" id="W5M0MpCehiHzreSzNTczkc9d"?>
<x:xmpmeta xmlns:x="adobe:ns:meta/">
  <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#">
    <rdf:Description xmlns:pdfaid="http://www.aiim.org/pdfa/ns/id/" rdf:about="">
      <pdfaid:part>3</pdfaid:part>
      <pdfaid:conformance>B</pdfaid:conformance>
    </rdf:Description>
    <rdf:Description xmlns:dc="http://purl.org/dc/elements/1.1/" rdf:about="">
      <dc:title>
        <rdf:Alt>
          <rdf:li xml:lang="x-default">{title}</rdf:li>
        </rdf:Alt>
      </dc:title>
      <dc:creator>
        <rdf:Seq>
          <rdf:li>{author}</rdf:li>
        </rdf:Seq>
      </dc:creator>
      <dc:description>
        <rdf:Alt>
          <rdf:li xml:lang="x-default">{subject}</rdf:li>
        </rdf:Alt>
      </dc:description>
    </rdf:Description>
    <rdf:Description xmlns:pdf="http://ns.adobe.com/pdf/1.3/" rdf:about="">
      <pdf:Producer>{producer}</pdf:Producer>
    </rdf:Description>
    <rdf:Description xmlns:xmp="http://ns.adobe.com/xap/1.0/" rdf:about="">
      <xmp:CreatorTool>{creator_tool}</xmp:CreatorTool>
      <xmp:CreateDate>{timestamp}</xmp:CreateDate>
      <xmp:ModifyDate>{timestamp}</xmp:ModifyDate>
    </rdf:Description>
    <rdf:Description xmlns:pdfaExtension="http://www.aiim.org/pdfa/ns/extension/" xmlns:pdfaSchema="http://www.aiim.org/pdfa/ns/schema#" xmlns:pdfaProperty="http://www.aiim.org/pdfa/ns/property#" rdf:about="">
      <pdfaExtension:schemas>
        <rdf:Bag>
          <rdf:li rdf:parseType="Resource">
            <pdfaSchema:schema>Factur-X PDFA Extension Schema</pdfaSchema:schema>
            <pdfaSchema:namespaceURI>{urn}</pdfaSchema:namespaceURI>
            <pdfaSchema:prefix>fx</pdfaSchema:prefix>
            <pdfaSchema:property>
              <rdf:Seq>
                <rdf:li rdf:parseType="Resource">
                  <pdfaProperty:name>DocumentFileName</pdfaProperty:name>
                  <pdfaProperty:valueType>Text</pdfaProperty:valueType>
                  <pdfaProperty:category>external</pdfaProperty:category>
                  <pdfaProperty:description>The name of the embedded XML document</pdfaProperty:description>
                </rdf:li>
                <rdf:li rdf:parseType="Resource">
                  <pdfaProperty:name>DocumentType</pdfaProperty:name>
                  <pdfaProperty:valueType>Text</pdfaProperty:valueType>
                  <pdfaProperty:category>external</pdfaProperty:category>
                  <pdfaProperty:description>The type of the hybrid document in capital letters, e.g. INVOICE or ORDER</pdfaProperty:description>
                </rdf:li>
                <rdf:li rdf:parseType="Resource">
                  <pdfaProperty:name>Version</pdfaProperty:name>
                  <pdfaProperty:valueType>Text</pdfaProperty:valueType>
                  <pdfaProperty:category>external</pdfaProperty:category>
                  <pdfaProperty:description>The actual version of the standard applying to the embedded XML document</pdfaProperty:description>
                </rdf:li>
                <rdf:li rdf:parseType="Resource">
                  <pdfaProperty:name>ConformanceLevel</pdfaProperty:name>
                  <pdfaProperty:valueType>Text</pdfaProperty:valueType>
                  <pdfaProperty:category>external</pdfaProperty:category>
                  <pdfaProperty:description>The conformance level of the embedded XML document</pdfaProperty:description>
                </rdf:li>
              </rdf:Seq>
            </pdfaSchema:property>
          </rdf:li>
        </rdf:Bag>
      </pdfaExtension:schemas>
    </rdf:Description>
    <rdf:Description xmlns:fx="{urn}" rdf:about="">
      <fx:DocumentFileName>{xml_filename}</fx:DocumentFileName>
      <fx:DocumentType>{documenttype}</fx:DocumentType>
      <fx:Version>{version}</fx:Version>
      <fx:ConformanceLevel>{xmp_level}</fx:ConformanceLevel>
    </rdf:Description>
  </rdf:RDF>
</x:xmpmeta>
<?xpacket end="w"?>
"""



import lxml.etree
import os
import pytest
from difflib import unified_diff
from xml.dom import minidom

from drafthorse.models.document import Document
from drafthorse.utils import validate_xml

samples = [
    f
    for f in os.listdir(os.path.join(os.path.dirname(__file__), "samples"))
    if f.endswith(".xml")
]


def diff_xml(a, b):
    for line in unified_diff(a.splitlines(), b.splitlines()):
        print(line)


def prettify(xml, **kwargs):
    try:
        from lxml import etree
    except ImportError:
        reparsed = minidom.parseString(xml)
        return reparsed.toprettyxml(indent="\t")
    else:
        parser = etree.XMLParser(remove_blank_text=True, **kwargs)
        return etree.tostring(etree.fromstring(xml, parser), pretty_print=True)


@pytest.mark.parametrize("filename", samples)
def test_sample_roundtrip(filename):
    origxml = prettify(
        open(os.path.join(os.path.dirname(__file__), "samples", filename), "rb").read(),
        remove_comments=True,
    )
    if filename.split("_")[2] != "XRECHNUNG":
        schema = "FACTUR-X_" + filename.split("_")[2]
    else:
        schema = "FACTUR-X_EN16931"

    # Validate that the sample file is valid, otherwise the test is moot
    validate_xml(xmlout=origxml, schema=schema)

    # Parse the sample file into our internal python structure
    doc = Document.parse(origxml)

    # Validate output XML and render a diff for debugging
    # skip first line (namespace order…)
    origxml = b"\n".join(origxml.split(b"\n")[1:]).decode().strip()
    try:
        generatedxml = prettify(doc.serialize(schema))
        generatedxml = b"\n".join(generatedxml.split(b"\n")[1:]).decode().strip()
        diff_xml(origxml, generatedxml)
    except lxml.etree.XMLSyntaxError:
        generatedxml = prettify(doc.serialize(None))
        generatedxml = b"\n".join(generatedxml.split(b"\n")[1:]).decode().strip()
        diff_xml(origxml, generatedxml)
        raise

    # Compare output XML
    assert origxml == generatedxml



import xml.etree.cElementTree as ET

from .container import StringContainer, Container
from .note import IncludedNote

from . import BASIC, EXTENDED, NS_QDT, NS_RAM, NS_RSM, NS_UDT
from .elements import Element
from .fields import (
    DateTimeField,
    Field,
    IndicatorField,
    MultiField,
    MultiStringField,
    StringField,
)
from .trade import TradeTransaction


class GuidelineDocumentContextParameter(Element):
    id = StringField(NS_RAM, "ID")

    class Meta:
        namespace = NS_RAM
        tag = "GuidelineSpecifiedDocumentContextParameter"


class BusinessDocumentContextParameter(Element):
    id = StringField(NS_RAM, "ID")

    class Meta:
        namespace = NS_RAM
        tag = "BusinessProcessSpecifiedDocumentContextParameter"


class DocumentContext(Element):
    test_indicator = IndicatorField(
        NS_RAM, "TestIndicator", required=False, profile=EXTENDED, _d="Testkennzeichen"
    )
    business_parameter: BusinessDocumentContextParameter = Field(
        BusinessDocumentContextParameter,
        required=False,
        profile=EXTENDED,
        _d="Geschäftsprozess, Wert",
    )
    guideline_parameter: GuidelineDocumentContextParameter = Field(
        GuidelineDocumentContextParameter,
        required=True,
        profile=BASIC,
        _d="Anwendungsempfehlung",
    )

    class Meta:
        namespace = NS_RSM
        tag = "ExchangedDocumentContext"


class EffectivePeriod(Element):
    complete = DateTimeField(NS_RAM, "CompleteDateTime")

    class Meta:
        namespace = NS_RAM
        tag = "EffectiveSpecifiedPeriod"


class Header(Element):
    id = StringField(NS_RAM, "ID", required=True, profile=BASIC, _d="Rechnungsnummer")
    name = StringField(
        NS_RAM, "Name", required=True, profile=BASIC, _d="Dokumentenart (Freitext)"
    )
    type_code = StringField(
        NS_RAM, "TypeCode", required=True, profile=BASIC, _d="Dokumentenart (Code)"
    )
    issue_date_time = DateTimeField(
        NS_RAM, "IssueDateTime", required=True, profile=BASIC, _d="Rechnungsdatum"
    )
    copy_indicator = IndicatorField(
        NS_RAM,
        "CopyIndicator",
        required=False,
        profile=EXTENDED,
        _d="Indikator Original/Kopie",
    )
    languages: StringContainer = MultiStringField(
        NS_RAM, "LanguageID", required=False, profile=EXTENDED
    )
    notes: Container = MultiField(IncludedNote)
    effective_period = Field(
        EffectivePeriod,
        required=False,
        profile=EXTENDED,
        _d="Vertragliches Fälligkeitsdatum der Rechnung",
    )

    class Meta:
        namespace = NS_RSM
        tag = "ExchangedDocument"


class Document(Element):
    context: DocumentContext = Field(DocumentContext, required=True)
    header: Header = Field(Header, required=True)
    trade: TradeTransaction = Field(TradeTransaction, required=True)
    __namespaces = {
        "rsm": NS_RSM,
        "qdt": NS_QDT,
        "ram": NS_RAM,
        "xs": "http://www.w3.org/2001/XMLSchema",
        "xsi": "http://www.w3.org/2001/XMLSchema-instance",
        "udt": NS_UDT,
    }

    def __init__(self):
        super().__init__()
        for ns, url in self.__namespaces.items():
            ET.register_namespace(ns, url)

    def serialize(self, schema="FACTUR-X_BASIC"):
        # First pass
        xml = super().serialize(schema)

        # Second pass to ensure all namespaces are defined even if they are unused
        root = ET.fromstring(xml)
        for ns, url in self.__namespaces.items():
            if ns.encode() not in xml:
                root.set(f"xmlns:{ns}", url)
        return ET.tostring(root, "utf-8")

    class Meta:
        namespace = NS_RSM
        tag = "CrossIndustryInvoice"



NS_RSM = "urn:un:unece:uncefact:data:standard:CrossIndustryInvoice:100"
NS_UDT = "urn:un:unece:uncefact:data:standard:UnqualifiedDataType:100"
NS_RAM = (
    "urn:un:unece:uncefact:data:standard:ReusableAggregateBusinessInformationEntity:100"
)
NS_QDT = "urn:un:unece:uncefact:data:standard:QualifiedDataType:100"
BASIC = "BASIC"
COMFORT = "COMFORT"
EXTENDED = "EXTENDED"



# Based on code of the factur-x Python package:
# Copyright 2016-2018, Alexis de Lattre <alexis.delattre@akretion.com>
# Modified for ZUGFeRD and Python 3 by Raphael Michel, 2018
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * The name of the authors may not be used to endorse or promote products
#      derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import logging
from datetime import datetime, timezone
from io import BytesIO
from lxml import etree
from pypdf import PdfReader, PdfWriter
from pypdf.generic import (
    ArrayObject,
    DecodedStreamObject,
    DictionaryObject,
    NameObject,
    create_string_object,
)

from drafthorse.xmp_schema import XMP_SCHEMA

logging.basicConfig()
logger = logging.getLogger("drafthorse")
logger.setLevel(logging.INFO)


def attach_xml(original_pdf, xml_data, level=None, metadata=None, lang=None):
    """
    Create the ZUGFeRD invoice by attaching
    the input XML and proper metadata
    :param original_pdf: Input PDF
    :param xml_data: Input XML
    :param level: optional Factur-X profile level
    one of ``{'MINIMUM', 'BASIC WL', 'BASIC', 'EN 16931', 'EXTENDED', 'XRECHNUNG'}``.
    If omitted, autodetection is performed
    :type level: string
    :param metadata: optional dict with user defined PDF metadata
    for fields "author", "keywords", "title", "subject", "creator" and "producer". If metadata is None (default value),
    this lib will generate some metadata by extracting relevant info from the Factur-X/Order-X XML.
    Here is an example for the metadata argument:
    ```
    pdf_metadata = {
        'author': 'MyCompany',
        'keywords': 'Factur-X, Invoice',
        'title': 'MyCompany: Invoice I1242',
        'creator': 'My Company Application',
        'producer': 'My company Accountant',
        'subject':
          'Factur-X invoice I1242 dated 2017-08-17 issued by MyCompany',
    }
    ```
    :type metadata: dict
    :param lang: Language identifier in RFC 3066 format to specify the
    natural language of the PDF document. Used by PDF readers for blind people.
    Example: en-US or fr-FR
    :type lang: string
    :return: Output PDF containing the metadata and XML
    """
    if not isinstance(original_pdf, bytes):
        raise TypeError("Please supply original PDF as bytes.")
    if not isinstance(xml_data, bytes):
        raise TypeError("Please supply XML data as bytes.")

    reader = PdfReader(BytesIO(original_pdf))
    output = PdfWriter()

    output._header = "%PDF-1.6\r\n%\xc7\xec\x8f\xa2".encode()
    output.append_pages_from_reader(reader)

    original_pdf_id = reader.trailer.get("/ID")
    if original_pdf_id:
        output._ID = original_pdf_id
        # else : generate some ?

    # Extract metadata from XML
    pdf_metadata, profile = _extract_xml_info(xml_data, level, metadata)

    # Extract output intents from input PDF
    output_intents = _get_original_output_intents(reader)

    _update_metadata_add_attachment(
        output, xml_data, pdf_metadata, profile, output_intents, lang
    )

    outbuffer = BytesIO()
    output.write(outbuffer)
    outbuffer.seek(0)
    return outbuffer.read()


def _get_original_output_intents(original_pdf):
    """
    Get output intents from input PDF
    :param original_pdf: Input PDF
    :return: Output PDF metadata information
    """
    output_intents = []
    try:
        pdf_root = original_pdf.trailer["/Root"]
        ori_output_intents = pdf_root["/OutputIntents"]
        for ori_output_intent in ori_output_intents:
            ori_output_intent_dict = ori_output_intent.get_object()
            dest_output_profile_dict = ori_output_intent_dict[
                "/DestOutputProfile"
            ].get_object()
            output_intents.append((ori_output_intent_dict, dest_output_profile_dict))
    except KeyError as ex:
        logger.debug("Data missing from PDF: %s", ex)
    except Exception as ex:
        logger.exception(ex)
    return output_intents


def _prepare_pdf_metadata_txt(pdf_metadata):
    """
    Create PDF info for the Document Properties section
    :param pdf_metadata: Metadata
    :return: PDF info
    """
    pdf_date = datetime.now(tz=timezone.utc).strftime("D:%Y%m%d%H%M%SZ")
    return {
        "/Author": pdf_metadata.get("author", ""),
        "/CreationDate": pdf_date,
        "/Creator": pdf_metadata.get("creator", "python-drafthorse"),
        "/Keywords": pdf_metadata.get("keywords", ""),
        "/ModDate": pdf_date,
        "/Subject": pdf_metadata.get("subject", ""),
        "/Title": pdf_metadata.get("title", ""),
    }


def _prepare_xmp_metadata(profile, pdf_metadata):
    """
    Prepare pdf metadata using the FACTUR-X XMP extension schema
    :param profile: Invoice profile
    :param pdf_metadata: PDF metadata
    :return: metadata XML
    """
    xml_str = XMP_SCHEMA.format(
        title=pdf_metadata.get("title", ""),
        author=pdf_metadata.get("author", ""),
        subject=pdf_metadata.get("subject", ""),
        producer=pdf_metadata.get("producer", "pypdf"),
        creator_tool=pdf_metadata.get("creator", "python-drafthorse"),
        timestamp=datetime.now(tz=timezone.utc).strftime("%Y-%m-%dT%H:%M:%S+00:00"),
        urn="urn:factur-x:pdfa:CrossIndustryDocument:invoice:1p0#",
        documenttype="INVOICE",
        xml_filename="factur-x.xml",
        version="1.0",
        xmp_level=profile,
    )
    return xml_str.encode("utf-8")


def _update_metadata_add_attachment(
    pdf_filestream,
    facturx_xml_str,
    pdf_metadata,
    facturx_level,
    output_intents,
    lang=None,
):
    """
    Update PDF metadata and attach XML file
    :param pdf_filestream: PDF data
    :param facturx_xml_str: XML data
    :param pdf_metadata: PDF metadata
    :param facturx_level: Invoice profile
    :param output_intents: Output intents from input PDF
    :param lang: Language identifier in RFC 3066 format
    """
    # Disable encoding
    # md5sum = hashlib.md5(facturx_xml_str).hexdigest()
    # md5sum_obj = create_string_object(md5sum)
    pdf_date = datetime.now(tz=timezone.utc).strftime("D:%Y%m%d%H%M%SZ")
    params_dict = DictionaryObject(
        {
            # NameObject('/CheckSum'): md5sum_obj,
            NameObject("/ModDate"): create_string_object(pdf_date),
            NameObject("/CreationDate"): create_string_object(pdf_date),
            NameObject("/Size"): create_string_object(str(len(facturx_xml_str))),
        }
    )
    file_entry = DecodedStreamObject()
    file_entry.set_data(facturx_xml_str)
    file_entry.update(
        {
            NameObject("/Type"): NameObject("/EmbeddedFile"),
            NameObject("/Params"): params_dict,
            NameObject("/Subtype"): NameObject("/text/xml"),
        }
    )
    file_entry_obj = pdf_filestream._add_object(file_entry)
    # The Filespec entry
    ef_dict = DictionaryObject(
        {NameObject("/F"): file_entry_obj, NameObject("/UF"): file_entry_obj}
    )

    fname_obj = create_string_object("factur-x.xml")
    filespec_dict = DictionaryObject(
        {
            NameObject("/AFRelationship"): NameObject(
                "/Data" if facturx_level in ("BASIC-WL", "MINIMUM") else "/Alternative"
            ),
            NameObject("/Desc"): create_string_object(
                "Invoice metadata conforming to ZUGFeRD standard (https://www.ferd-net.de/)"
            ),
            NameObject("/Type"): NameObject("/Filespec"),
            NameObject("/F"): fname_obj,
            NameObject("/EF"): ef_dict,
            NameObject("/UF"): fname_obj,
        }
    )
    filespec_obj = pdf_filestream._add_object(filespec_dict)
    name_arrayobj_cdict = {fname_obj: filespec_obj}
    name_arrayobj_content_sort = sorted(name_arrayobj_cdict.items(), key=lambda x: x[0])
    name_arrayobj_content_final = []
    af_list = []
    for fname_obj, filespec_obj in name_arrayobj_content_sort:
        name_arrayobj_content_final += [fname_obj, filespec_obj]
        af_list.append(filespec_obj)
    embedded_files_names_dict = DictionaryObject(
        {NameObject("/Names"): ArrayObject(name_arrayobj_content_final)}
    )
    # Then create the entry for the root, as it needs a
    # reference to the Filespec
    embedded_files_dict = DictionaryObject(
        {NameObject("/EmbeddedFiles"): embedded_files_names_dict}
    )
    res_output_intents = []
    for output_intent_dict, dest_output_profile_dict in output_intents:
        dest_output_profile_obj = pdf_filestream._add_object(dest_output_profile_dict)
        # TODO detect if there are no other objects in output_intent_dest_obj
        # than /DestOutputProfile
        output_intent_dict.update(
            {NameObject("/DestOutputProfile"): dest_output_profile_obj}
        )
        output_intent_obj = pdf_filestream._add_object(output_intent_dict)
        res_output_intents.append(output_intent_obj)
    # Update the root
    metadata_xml_str = _prepare_xmp_metadata(facturx_level, pdf_metadata)
    metadata_file_entry = DecodedStreamObject()
    metadata_file_entry.set_data(metadata_xml_str)
    metadata_file_entry.update(
        {
            NameObject("/Subtype"): NameObject("/XML"),
            NameObject("/Type"): NameObject("/Metadata"),
        }
    )
    metadata_obj = pdf_filestream._add_object(metadata_file_entry)
    af_value_obj = pdf_filestream._add_object(ArrayObject(af_list))
    pdf_filestream._root_object.update(
        {
            NameObject("/AF"): af_value_obj,
            NameObject("/Metadata"): metadata_obj,
            NameObject("/Names"): embedded_files_dict,
            # show attachments when opening PDF
            NameObject("/PageMode"): NameObject("/UseAttachments"),
        }
    )
    if res_output_intents:
        pdf_filestream._root_object.update(
            {NameObject("/OutputIntents"): ArrayObject(res_output_intents)}
        )
    if lang:
        pdf_filestream._root_object.update(
            {
                NameObject("/Lang"): create_string_object(lang.replace("_", "-")),
            }
        )
    metadata_txt_dict = _prepare_pdf_metadata_txt(pdf_metadata)
    pdf_filestream.add_metadata(metadata_txt_dict)


def _extract_xml_info(xml_data, level=None, metadata=None):
    """
    Extract metadata and profile from XML further added to the PDF
    :param xml_data: XML data
    :param level: optional Factur-X profile level
       one of {MINIMUM, BASIC WL, BASIC, EN 16931, EXTENDED, XRECHNUNG}
       if omitted autodetection is performed
    :param metadata: optional dict with user defined pdf_metadata
        for fields "author", "keywords", "title" and "subject"
    :return: Metadata and profile
    """

    xml_etree = etree.fromstring(xml_data)
    namespaces = xml_etree.nsmap

    # get metadata
    number_xpath = xml_etree.xpath(
        "//rsm:ExchangedDocument/ram:ID", namespaces=namespaces
    )
    number = number_xpath[0].text
    seller_xpath = xml_etree.xpath(
        "//ram:ApplicableHeaderTradeAgreement/ram:SellerTradeParty/ram:Name",
        namespaces=namespaces,
    )
    seller = seller_xpath[0].text

    if metadata is None:
        metadata = {}
    pdf_metadata = {
        "author": metadata.get("author", seller),
        "keywords": metadata.get("keywords", "Factur-X"),
        "title": metadata.get("title", number),
        "subject": metadata.get("subject", number),
        "producer": metadata.get("producer", "pypdf"),
        "creator": metadata.get("creator", "python-drafthorse"),
    }

    # get profile
    doc_id_xpath = xml_etree.xpath(
        "//rsm:ExchangedDocumentContext"
        "/ram:GuidelineSpecifiedDocumentContextParameter"
        "/ram:ID",
        namespaces=namespaces,
    )
    doc_id = doc_id_xpath[0].text

    if level is None:
        # autodetection of Factur-X profile
        profile = doc_id.split(":")[-1]
        if doc_id.split(":")[-1] in ["basic", "extended"]:
            profile = doc_id.split(":")[-1]
        elif doc_id.split(":")[-1].startswith("xrechnung"):
            profile = "xrechnung"
        elif doc_id.split(":")[-2] == "en16931":
            profile = doc_id.split(":")[-2]
            profile = profile[:2] + " " + profile[2:]
        else:
            raise Exception("Invalid XML profile!")
    else:
        profile = level

    profile = profile.upper()
    logger.info("Factur-X profile detected from XML: %s", profile)

    return pdf_metadata, profile



import collections
import sys
import xml.etree.cElementTree as ET
from collections import OrderedDict
from datetime import datetime
from decimal import Decimal

from drafthorse.utils import validate_xml

from . import NS_UDT
from .container import Container
from .fields import Field


class BaseElementMeta(type):
    @classmethod
    def __prepare__(self, name, bases):
        del name, bases
        return collections.OrderedDict()

    def __new__(mcls, name, bases, attrs):
        cls = super(BaseElementMeta, mcls).__new__(mcls, name, bases, attrs)
        fields = list(cls._fields) if hasattr(cls, "_fields") else []
        for attr, obj in attrs.items():
            if isinstance(obj, Field):
                if sys.version_info < (3, 6):
                    obj.__set_name__(cls, attr)
                fields.append(obj)
        cls._fields = fields
        return cls


class Element(metaclass=BaseElementMeta):
    def __init__(self, **kwargs):
        self.required = kwargs.get("required", False)
        self._data = OrderedDict(
            [
                (f.name, f.initialize() if f.default or f.required else None)
                for f in self._fields
            ]
        )
        for k, v in kwargs.items():
            setattr(self, k, v)

    def _etree_node(self):
        node = ET.Element(self.get_tag())
        if hasattr(self, "Meta") and hasattr(self.Meta, "attributes"):
            for k, v in self.Meta.attributes.items():
                node.set(k, v)
        return node

    def to_etree(self):
        node = self._etree_node()
        for _, v in self._data.items():
            if v is not None:
                v.append_to(node)
        return node

    def get_tag(self):
        return "{%s}%s" % (self.Meta.namespace, self.Meta.tag)

    def is_empty(self, el):
        return not list(el) and not el.text

    def append_to(self, node):
        el = self.to_etree()
        if self.required or not self.is_empty(el):
            node.append(el)

    def serialize(self, schema="FACTUR-X_BASIC"):
        """
        Create XML from ZUGFeRD data model
        :param schema: XML schema
        :return: ZUGFeRD XML
        """
        xml = ET.tostring(self.to_etree(), "utf-8")

        return validate_xml(xml, schema)

    def __setattr__(self, key, value):
        if (
            not hasattr(self, key)
            and not key.startswith("_")
            and key not in ("required",)
        ):
            raise AttributeError(
                f"Element {type(self)} has no attribute '{key}'. If you set it, it would not be included in the output."
            )
        return super().__setattr__(key, value)

    def from_etree(self, root, strict=True):
        if (
            hasattr(self, "Meta")
            and hasattr(self.Meta, "namespace")
            and root.tag != "{%s}%s" % (self.Meta.namespace, self.Meta.tag)
        ):
            raise TypeError(
                "Invalid XML, found tag {} where {} was expected".format(
                    root.tag, "{%s}%s" % (self.Meta.namespace, self.Meta.tag)
                )
            )
        field_index = {}
        for field in self._fields:
            element = getattr(self, field.name)
            field_index[element.get_tag()] = (field.name, element)
        for child in root:
            if child.tag == ET.Comment:
                continue
            if child.tag in field_index:
                name, _childel = field_index[child.tag]
                if isinstance(getattr(self, name), Container):
                    getattr(self, name).add_from_etree(child, strict)
                else:
                    getattr(self, name).from_etree(child, strict)
            elif strict:
                raise TypeError("Unknown element {}".format(child.tag))
        return self

    @classmethod
    def parse(cls, xmlinput, strict=True):
        from lxml import etree

        root = etree.fromstring(xmlinput)
        return cls().from_etree(root, strict)


class StringElement(Element):
    def __init__(self, namespace, tag, text=""):
        super().__init__()
        self._namespace = namespace
        self._tag = tag
        self._text = text
        self._set_on_input = False

    def __repr__(self):
        return "<{}: {}>".format(type(self).__name__, str(self))

    def __str__(self):
        return str(self._text)

    def is_empty(self, el):
        return super().is_empty(el) and not self._set_on_input

    def get_tag(self):
        return "{%s}%s" % (self._namespace, self._tag)

    def to_etree(self):
        node = self._etree_node()
        node.text = self._text
        return node

    def from_etree(self, root, strict=True):
        self._text = root.text
        self._set_on_input = True
        return self


class DecimalElement(StringElement):
    def __init__(self, namespace, tag, value=None):
        super().__init__(namespace, tag)
        self._value = value

    def to_etree(self):
        node = self._etree_node()
        node.text = str(self._value) if self._value is not None else ""
        return node

    def __str__(self):
        return self._value

    def from_etree(self, root, strict=True):
        self._value = Decimal(root.text)
        self._set_on_input = True
        return self


class QuantityElement(StringElement):
    def __init__(self, namespace, tag, amount="", unit_code=""):
        super().__init__(namespace, tag)
        self._amount = amount
        self._unit_code = unit_code

    def to_etree(self):
        node = self._etree_node()
        node.text = str(self._amount)
        node.attrib["unitCode"] = self._unit_code
        return node

    def __str__(self):
        return "{} {}".format(self._amount, self._unit_code)

    def from_etree(self, root, strict=True):
        self._amount = Decimal(root.text)
        self._unit_code = root.attrib["unitCode"]
        self._set_on_input = True
        return self


class CurrencyElement(StringElement):
    def __init__(self, namespace, tag, amount="", currency=None):
        super().__init__(namespace, tag)
        self._amount = amount
        self._currency = currency

    def to_etree(self):
        node = self._etree_node()
        node.text = str(self._amount)
        if self._currency is not None:
            node.attrib["currencyID"] = self._currency
        elif "currencyID" in node.attrib:
            del node.attrib["currencyID"]
        return node

    def from_etree(self, root, strict=True):
        self._amount = Decimal(root.text)
        self._currency = root.attrib.get("currencyID") or None
        self._set_on_input = True
        return self

    def __str__(self):
        return "{} {}".format(self._amount, self._currency)


class ClassificationElement(StringElement):
    def __init__(self, namespace, tag, text="", list_id="", list_version_id=""):
        super().__init__(namespace, tag)
        self._text = text
        self._list_id = list_id
        self._list_version_id = list_version_id

    def to_etree(self):
        node = self._etree_node()
        node.text = self._text
        node.attrib["listID"] = self._list_id
        node.attrib["listVersionID"] = self._list_version_id
        return node

    def from_etree(self, root, strict=True):
        self._text = Decimal(root.text)
        self._list_id = root.attrib["listID"]
        self._list_version_id = root.attrib["listVersionID"]
        self._set_on_input = True
        return self

    def __str__(self):
        return "{} ({} {})".format(self._text, self._list_id, self._list_version_id)


class BinaryObjectElement(StringElement):
    def __init__(self, namespace, tag, text="", filename="", mime_code=""):
        super().__init__(namespace, tag)
        self._mime_code = mime_code
        self._filename = filename
        self._text = text

    def to_etree(self):
        node = self._etree_node()
        node.attrib["mimeCode"] = self._mime_code
        node.attrib["filename"] = self._filename
        node.text = self._text
        return node

    def from_etree(self, root, strict=True):
        self._mime_code = root.attrib["mimeCode"]
        self._filename = root.attrib["filename"]
        self._text = root.text
        self._set_on_input = True
        return self

    def __str__(self):
        return "{} ({} {})".format(self._text, self._filename, self._mime_code)


class AgencyIDElement(StringElement):
    def __init__(self, namespace, tag, text="", scheme_id=""):
        super().__init__(namespace, tag)
        self._text = text
        self._scheme_id = scheme_id

    def to_etree(self):
        node = self._etree_node()
        node.text = self._text
        node.attrib["schemeAgencyID"] = self._scheme_id
        return node

    def from_etree(self, root, strict=True):
        self._text = root.text
        self._scheme_id = root.attrib["schemeAgencyID"]
        self._set_on_input = True
        return self

    def __str__(self):
        return "{} ({})".format(self._text, self._scheme_id)


class IDElement(StringElement):
    def __init__(self, namespace, tag, text="", scheme_id=""):
        super().__init__(namespace, tag)
        self._text = text
        self._scheme_id = scheme_id

    def to_etree(self):
        node = self._etree_node()
        node.text = self._text
        if self._scheme_id != "":
            node.attrib["schemeID"] = self._scheme_id
        return node

    def from_etree(self, root, strict=True):
        self._text = root.text
        try:
            self._scheme_id = root.attrib["schemeID"]
        except Exception:
            root.attrib["schemeID"] = ""
            self._scheme_id = root.attrib["schemeID"]
        self._set_on_input = True
        return self

    def __str__(self):
        return "{} ({})".format(self._text, self._scheme_id)


class DateTimeElement(StringElement):
    def __init__(
        self, namespace, tag, value=None, format="102", date_time_namespace=NS_UDT
    ):
        super().__init__(namespace, tag)
        self._value = value
        self._format = format
        self._date_time_namespace = date_time_namespace

    def to_etree(self):
        t = self._etree_node()
        node = ET.Element("{%s}%s" % (self._date_time_namespace, "DateTimeString"))
        if self._value:
            if self._format == "102":
                node.text = self._value.strftime("%Y%m%d")
            elif self._format == "616":
                if sys.version_info < (3, 6):
                    node.text = "{}{}".format(
                        self._value.isocalendar()[0], self._value.isocalendar()[1]
                    )
                else:
                    node.text = self._value.strftime("%G%V")
            node.attrib["format"] = self._format
            t.append(node)
        return t

    def from_etree(self, root, strict=True):
        if len(root) != 1:
            raise TypeError("Date containers should have one child")
        if root[0].tag != "{%s}%s" % (self._date_time_namespace, "DateTimeString"):
            if strict:
                raise TypeError("Tag %s not recognized" % root[0].tag)
            else:
                return self
        self._format = root[0].attrib["format"]
        if self._format == "102":
            self._value = datetime.strptime(root[0].text, "%Y%m%d").date()
        elif self._format == "616":
            if sys.version_info < (3, 6):
                from isoweek import Week

                w = Week(int(root[0].text[:4]), int(root[0].text[4:]))
                self._value = w.monday()
            else:
                self._value = datetime.strptime(root[0].text + "1", "%G%V%u").date()
        elif strict:
            raise TypeError(
                "Date format %s cannot be parsed" % root[0].attrib["format"]
            )
        self._set_on_input = True
        return self

    def __str__(self):
        return "{}".format(self._value)


class DirectDateTimeElement(StringElement):
    def __init__(self, namespace, tag, value=None):
        super().__init__(namespace, tag)
        self._value = value

    def to_etree(self):
        t = self._etree_node()
        if self._value:
            t.text = self._value.strftime("%Y-%m-%dT%H:%M:%S")
        return t

    def from_etree(self, root, strict=True):
        try:
            self._value = datetime.strptime(root.text, "%Y-%m-%dT%H:%M:%S").date()
        except Exception:
            self._value = ""
        self._set_on_input = True
        return self

    def __str__(self):
        return "{}".format(self._value)


class IndicatorElement(StringElement):
    def __init__(self, namespace, tag, value=None):
        super().__init__(namespace, tag)
        self._value = value

    def get_tag(self):
        return "{%s}%s" % (self._namespace, self._tag)

    def to_etree(self):
        t = self._etree_node()
        if self._value is None:
            return t
        node = ET.Element("{%s}%s" % (NS_UDT, "Indicator"))
        node.text = str(self._value).lower()
        t.append(node)
        return t

    def __str__(self):
        return "{}".format(self._value)

    def from_etree(self, root, strict=True):
        if len(root) != 1:
            raise TypeError("Indicator containers should have one child")
        if root[0].tag != "{%s}%s" % (NS_UDT, "Indicator"):
            raise TypeError("Tag %s not recognized" % root[0].tag)
        self._value = root[0].text == "true"
        self._set_on_input = True
        return self



from . import BASIC, COMFORT, EXTENDED, NS_RAM
from .elements import Element
from .fields import DateTimeField, Field, StringField
from .party import (
    ShipFromTradeParty,
    ShipToTradeParty,
    UltimateShipToTradeParty,
)
from .references import (
    DeliveryNoteReferencedDocument,
    DespatchAdviceReferencedDocument,
    ReceivingAdviceReferencedDocument,
)


class SupplyChainEvent(Element):
    occurrence = DateTimeField(
        NS_RAM,
        "OccurrenceDateTime",
        required=False,
        profile=BASIC,
        _d="Tatsächlicher Lieferungszeitpunkt",
    )

    class Meta:
        namespace = NS_RAM
        tag = "ActualDeliverySupplyChainEvent"


class LogisticsTransportMovement(Element):
    mode_code = StringField(NS_RAM, "ModeCode", required=False, profile=EXTENDED)

    class Meta:
        namespace = NS_RAM
        tag = "SpecifiedLogisticsTransportMovement"


class SupplyChainConsignment(Element):  # TODO: Deprecated?
    movement: LogisticsTransportMovement = Field(
        LogisticsTransportMovement, required=False, profile=EXTENDED
    )

    class Meta:
        namespace = NS_RAM
        tag = "RelatedSupplyChainConsignment"


class TradeDelivery(Element):
    consignment: SupplyChainConsignment = Field(
        SupplyChainConsignment,
        default=False,
        required=False,
        profile=EXTENDED,
        _d="Detailinformationen zur Konsignation oder Sendung",
    )
    ship_to: ShipToTradeParty = Field(ShipToTradeParty, required=False, profile=COMFORT)
    ultimate_ship_to: UltimateShipToTradeParty = Field(
        UltimateShipToTradeParty, required=False, profile=EXTENDED
    )
    ship_from: ShipFromTradeParty = Field(
        ShipFromTradeParty, required=False, profile=EXTENDED
    )
    event: SupplyChainEvent = Field(SupplyChainEvent, required=False, profile=BASIC)
    despatch_advice: DespatchAdviceReferencedDocument = Field(
        DespatchAdviceReferencedDocument, required=False, profile=COMFORT
    )
    delivery_note: DeliveryNoteReferencedDocument = Field(
        DeliveryNoteReferencedDocument, required=False, profile=EXTENDED
    )
    receiving_advice: ReceivingAdviceReferencedDocument = Field(
        ReceivingAdviceReferencedDocument, required=False, profile=COMFORT
    )

    class Meta:
        namespace = NS_RAM
        tag = "ApplicableHeaderTradeDelivery"



from . import BASIC, COMFORT, EXTENDED, NS_RAM, NS_RSM, NS_QDT
from .accounting import (
    ApplicableTradeTax,
    AppliedTradeTax,
    BillingSpecifiedPeriod,
    MonetarySummation,
    ReceivableAccountingAccount,
    TradeAllowanceCharge,
)
from .delivery import TradeDelivery
from .container import Container
from .elements import Element
from .fields import DateTimeField, DecimalField, Field, MultiField, StringField
from .party import (
    BuyerTradeParty,
    EndUserTradeParty,
    InvoiceeTradeParty,
    InvoicerTradeParty,
    PayeeTradeParty,
    PayerTradeParty,
    SellerTaxRepresentativeTradeParty,
    SellerTradeParty,
)
from .payment import (
    PaymentMeans,
    PaymentTerms,
    TaxApplicableTradeCurrencyExchange,
)
from .references import (
    AdditionalReferencedDocument,
    BuyerOrderReferencedDocument,
    SellerOrderReferencedDocument,
    ContractReferencedDocument,
    InvoiceReferencedDocument,
    InvoiceSpecifiedReferencedDocument,
    ProcuringProjectType,
    UltimateCustomerOrderReferencedDocument,
)
from .tradelines import LineItem


class DeliveryTerms(Element):
    type_code = StringField(
        NS_RAM,
        "DeliveryTypeCode",
        required=False,
        profile=EXTENDED,
        _d="Lieferbedingung (Code)",
    )

    class Meta:
        namespace = NS_RAM
        tag = "ApplicableTradeDeliveryTerms"


class TradeAgreement(Element):
    buyer_reference = StringField(
        NS_RAM,
        "BuyerReference",
        required=False,
        profile=COMFORT,
        _d="Referenz des Käufers",
    )
    seller: SellerTradeParty = Field(
        SellerTradeParty, required=True, _d="Detailinformationen zum Verkäufer"
    )
    buyer: BuyerTradeParty = Field(BuyerTradeParty, required=True)
    end_user: EndUserTradeParty = Field(
        EndUserTradeParty, required=False, _d="Abweichender Endverbraucher"
    )
    delivery_terms: DeliveryTerms = Field(
        DeliveryTerms, required=False, profile=EXTENDED
    )
    seller_order: SellerOrderReferencedDocument = Field(
        SellerOrderReferencedDocument, required=False, profile=COMFORT
    )
    buyer_order: BuyerOrderReferencedDocument = Field(
        BuyerOrderReferencedDocument, required=False
    )
    contract: ContractReferencedDocument = Field(
        ContractReferencedDocument, required=False, profile=COMFORT
    )
    additional_references: Container = MultiField(
        AdditionalReferencedDocument, required=False, profile=COMFORT
    )
    description = StringField(NS_RAM, "Description", required=False, profile=COMFORT)
    seller_tax_representative_party: SellerTaxRepresentativeTradeParty = Field(
        SellerTaxRepresentativeTradeParty, required=False
    )
    procuring_project_type: ProcuringProjectType = Field(
        ProcuringProjectType, required=False
    )
    customer_order: UltimateCustomerOrderReferencedDocument = Field(
        UltimateCustomerOrderReferencedDocument, required=False, profile=COMFORT
    )

    class Meta:
        namespace = NS_RAM
        tag = "ApplicableHeaderTradeAgreement"


class LogisticsServiceCharge(Element):
    description = StringField(
        NS_RAM,
        "Description",
        required=True,
        profile=COMFORT,
        _d="Identifikation der Servicegebühr",
    )
    applied_amount = DecimalField(
        NS_RAM,
        "AppliedAmount",
        required=True,
        profile=COMFORT,
        _d="Betrag der Servicegebühr",
    )
    trade_tax: Container = MultiField(AppliedTradeTax, required=False, profile=COMFORT)

    class Meta:
        namespace = NS_RAM
        tag = "SpecifiedLogisticsServiceCharge"


class IncludedTradeTax(Element):
    calculated_amount = DecimalField(
        NS_RAM, "CalculatedAmount", required=True, profile=BASIC, _d="Steuerbetrag"
    )
    type_code = StringField(
        NS_RAM, "TypeCode", required=True, profile=BASIC, _d="Steuerart (Code)"
    )
    exemption_reason = StringField(
        NS_RAM,
        "ExemptionReason",
        required=False,
        profile=COMFORT,
        _d="Grund der Steuerbefreiung (Freitext)",
    )
    exemption_reason_code = StringField(
        NS_RAM,
        "ExemptionReasonCode",
        required=False,
        profile=EXTENDED,
        _d="Grund der Steuerbefreiung (Code)",
    )
    category_code = StringField(
        NS_RAM,
        "CategoryCode",
        required=False,
        profile=COMFORT,
        _d="Steuerkategorie (Wert)",
    )
    rate_applicable_percent = DecimalField(
        NS_RAM, "RateApplicablePercent", required=True, profile=BASIC
    )

    class Meta:
        namespace = NS_RAM
        tag = "IncludedTradeTax"


class AdvancePayment(Element):
    paid_amount = DecimalField(NS_RAM, "PaidAmount")
    received_date = DateTimeField(
        NS_RAM, "FormattedReceivedDateTime", date_time_namespace=NS_QDT
    )
    included_trade_tax: Container = MultiField(IncludedTradeTax)
    referenced_document: InvoiceSpecifiedReferencedDocument = Field(
        InvoiceSpecifiedReferencedDocument, required=False, profile=EXTENDED
    )

    class Meta:
        namespace = NS_RAM
        tag = "SpecifiedAdvancePayment"


class TradeSettlement(Element):
    creditor_reference_id = StringField(NS_RAM, "CreditorReferenceID")
    payment_reference = StringField(NS_RAM, "PaymentReference")
    tax_currency_code = StringField(
        NS_RAM, "TaxCurrencyCode", required=False, profile=BASIC
    )
    currency_code = StringField(NS_RAM, "InvoiceCurrencyCode")
    issuer_reference = StringField(NS_RAM, "InvoiceIssuerReference", profile=EXTENDED)
    invoicer: InvoicerTradeParty = Field(
        InvoicerTradeParty, required=False, profile=EXTENDED, _d="Rechnungsaussteller"
    )
    invoicee: InvoiceeTradeParty = Field(
        InvoiceeTradeParty, required=False, profile=EXTENDED, _d="Rechnungsempfänger"
    )
    payee: PayeeTradeParty = Field(
        PayeeTradeParty, required=False, profile=BASIC, _d="Zahlungsempfänger"
    )
    payer: PayerTradeParty = Field(
        PayerTradeParty, required=False, profile=EXTENDED, _d="Zahlungspflichtiger"
    )
    invoice_currency: TaxApplicableTradeCurrencyExchange = Field(
        TaxApplicableTradeCurrencyExchange, profile=EXTENDED
    )
    payment_means: PaymentMeans = Field(PaymentMeans)
    trade_tax: Container = MultiField(ApplicableTradeTax)
    period: BillingSpecifiedPeriod = Field(
        BillingSpecifiedPeriod, required=False, profile=BASIC
    )
    allowance_charge: Container = MultiField(
        TradeAllowanceCharge,
        required=False,
        profile=BASIC,
        _d="Schalter für Zu-/Abschlag",
    )
    service_charge: Container = MultiField(
        LogisticsServiceCharge, required=False, profile=EXTENDED
    )
    terms: Container = MultiField(PaymentTerms, required=False, profile=BASIC)
    monetary_summation: MonetarySummation = Field(
        MonetarySummation,
        required=True,
        profile=BASIC,
        _d="Detailinformation zu Belegsummen",
    )
    invoice_referenced_document: InvoiceReferencedDocument = Field(
        InvoiceReferencedDocument, required=False, profile=BASIC
    )
    accounting_account: ReceivableAccountingAccount = Field(
        ReceivableAccountingAccount,
        required=False,
        profile=BASIC,
        _d="Detailinformationen zur Buchungsreferenz",
    )
    advance_payment: Container = MultiField(
        AdvancePayment, required=False, profile=EXTENDED
    )

    class Meta:
        namespace = NS_RAM
        tag = "ApplicableHeaderTradeSettlement"


class TradeTransaction(Element):
    items: Container = MultiField(LineItem, required=True)
    agreement: TradeAgreement = Field(TradeAgreement, required=True)
    delivery: TradeDelivery = Field(TradeDelivery, required=True)
    settlement: TradeSettlement = Field(TradeSettlement, required=True)

    class Meta:
        namespace = NS_RSM
        tag = "SupplyChainTradeTransaction"



from .container import CurrencyContainer, Container
from . import BASIC, COMFORT, EXTENDED, NS_RAM
from .elements import Element
from .fields import (
    CurrencyField,
    DateTimeField,
    DecimalField,
    IndicatorField,
    MultiCurrencyField,
    MultiField,
    QuantityField,
    StringField,
)


class BillingSpecifiedPeriod(Element):
    description = StringField(
        NS_RAM,
        "Description",
        required=True,
        profile=COMFORT,
        _d="Freitext der Zahlungsbedingungen",
    )
    start = DateTimeField(NS_RAM, "StartDateTime", required=True, profile=COMFORT)
    end = DateTimeField(NS_RAM, "EndDateTime", required=True, profile=COMFORT)

    class Meta:
        namespace = NS_RAM
        tag = "BillingSpecifiedPeriod"


class ApplicableTradeTax(Element):
    calculated_amount = DecimalField(
        NS_RAM, "CalculatedAmount", required=True, profile=BASIC, _d="Steuerbetrag"
    )
    type_code = StringField(
        NS_RAM, "TypeCode", required=True, profile=BASIC, _d="Steuerart (Code)"
    )
    exemption_reason = StringField(
        NS_RAM,
        "ExemptionReason",
        required=False,
        profile=BASIC,
        _d="Grund der Steuerbefreiung (Freitext)",
    )
    basis_amount = DecimalField(
        NS_RAM,
        "BasisAmount",
        required=True,
        profile=BASIC,
        _d="Basisbetrag der Steuerberechnung",
    )
    line_total_basis_amount = DecimalField(
        NS_RAM,
        "LineTotalBasisAmount",
        required=False,
        profile=EXTENDED,
        _d="Warenbetrag des Steuersatzes",
    )
    allowance_charge_basis_amount = DecimalField(
        NS_RAM,
        "AllowanceChargeBasisAmount",
        required=False,
        profile=EXTENDED,
        _d="Gesamtbetrag Zu- und Abschläge des Steuersatzes",
    )
    category_code = StringField(
        NS_RAM,
        "CategoryCode",
        required=False,
        profile=BASIC,
        _d="Steuerkategorie (Wert)",
    )
    exemption_reason_code = StringField(
        NS_RAM,
        "ExemptionReasonCode",
        required=False,
        profile=BASIC,
        _d="Grund der Steuerbefreiung (Code)",
    )
    tax_point_date = DateTimeField(
        NS_RAM, "TaxPointDate", required=False, profile=COMFORT
    )
    due_date_type_code = StringField(
        NS_RAM,
        "DueDateTypeCode",
        required=False,
        profile=BASIC,
    )
    rate_applicable_percent = DecimalField(
        NS_RAM, "RateApplicablePercent", required=True, profile=BASIC
    )

    class Meta:
        namespace = NS_RAM
        tag = "ApplicableTradeTax"


class AccountingAccount(Element):
    id = StringField(
        NS_RAM, "ID", required=True, profile=EXTENDED, _d="Buchungsreferenz"
    )

    class Meta:
        namespace = NS_RAM
        tag = "SpecifiedTradeAccountingAccount"


class ReceivableAccountingAccount(Element):
    id = StringField(
        NS_RAM, "ID", required=True, profile=EXTENDED, _d="Buchungsreferenz"
    )
    type_code = StringField(NS_RAM, "TypeCode", required=True, profile=EXTENDED)

    class Meta:
        namespace = NS_RAM
        tag = "ReceivableSpecifiedTradeAccountingAccount"


class MonetarySummation(Element):
    line_total = DecimalField(
        NS_RAM,
        "LineTotalAmount",
        required=True,
        profile=BASIC,
        _d="Gesamtbetrag der Positionen",
    )
    charge_total = DecimalField(
        NS_RAM,
        "ChargeTotalAmount",
        required=True,
        profile=BASIC,
        _d="Gesamtbetrag der Zuschläge",
    )
    allowance_total = DecimalField(
        NS_RAM,
        "AllowanceTotalAmount",
        required=True,
        profile=BASIC,
        _d="Gesamtbetrag der Abschläge",
    )
    tax_basis_total = CurrencyField(
        NS_RAM,
        "TaxBasisTotalAmount",
        required=True,
        profile=BASIC,
        _d="Steuerbasisbetrag",
    )
    tax_total = CurrencyField(
        NS_RAM, "TaxTotalAmount", required=True, profile=BASIC, _d="Steuergesamtbetrag"
    )
    tax_total_other_currency: CurrencyContainer = MultiCurrencyField(
        NS_RAM, "TaxTotalAmount", profile=EXTENDED, _d="Steuergesamtbetrag"
    )
    rounding_amount = DecimalField(
        NS_RAM,
        "RoundingAmount",
        required=False,
        profile=COMFORT,
    )
    grand_total = CurrencyField(
        NS_RAM, "GrandTotalAmount", required=True, profile=BASIC, _d="Bruttosumme"
    )
    prepaid_total = DecimalField(
        NS_RAM,
        "TotalPrepaidAmount",
        required=False,
        profile=COMFORT,
        _d="Anzahlungsbetrag",
    )
    due_amount = DecimalField(
        NS_RAM, "DuePayableAmount", required=False, profile=COMFORT, _d="Zahlbetrag"
    )

    class Meta:
        namespace = NS_RAM
        tag = "SpecifiedTradeSettlementHeaderMonetarySummation"


class AppliedTradeTax(Element):
    type_code = StringField(NS_RAM, "TypeCode", required=True, profile=COMFORT)
    category_code = StringField(NS_RAM, "CategoryCode", required=True, profile=COMFORT)
    rate_applicable_percent = DecimalField(
        NS_RAM, "RateApplicablePercent", required=True, profile=COMFORT
    )

    class Meta:
        namespace = NS_RAM
        tag = "AppliedTradeTax"


class CategoryTradeTax(AppliedTradeTax):
    class Meta:
        namespace = NS_RAM
        tag = "CategoryTradeTax"


class TradeAllowanceCharge(Element):
    indicator = IndicatorField(
        NS_RAM,
        "ChargeIndicator",
        required=False,
        profile=COMFORT,
        _d="Schalter für Zu-/Abschlag",
    )
    sequence_numeric = DecimalField(
        NS_RAM,
        "SequenceNumeric",
        required=False,
        profile=EXTENDED,
        _d="Berechnungsreihenfolge",
    )
    calculation_percent = DecimalField(  # TODO: Should be deprecated?
        NS_RAM,
        "CalculationPercent",
        required=False,
        profile=COMFORT,
        _d="Rabatt in Prozent",
    )
    basis_amount = DecimalField(  # TODO: Should be deprecated?
        NS_RAM,
        "BasisAmount",
        required=False,
        profile=COMFORT,
        _d="Basisbetrag des Rabatts",
    )
    basis_quantity = QuantityField(
        NS_RAM,
        "BasisQuantity",
        required=False,
        profile=EXTENDED,
        _d="Basismenge des Rabatts",
    )
    actual_amount = DecimalField(
        NS_RAM,
        "ActualAmount",
        required=True,
        profile=COMFORT,
        _d="Betrag des Zu-/Abschlags",
    )
    reason_code = StringField(NS_RAM, "ReasonCode", required=False, profile=COMFORT)
    reason = StringField(NS_RAM, "Reason", required=False, profile=COMFORT)
    trade_tax: Container = MultiField(CategoryTradeTax, required=False, profile=COMFORT)

    class Meta:
        namespace = NS_RAM
        tag = "SpecifiedTradeAllowanceCharge"



import logging
import os

logger = logging.getLogger("drafthorse")


def validate_xml(xmlout, schema):
    try:
        from lxml import etree
    except ImportError:
        logger.warning("Could not validate output as LXML is not installed.")
        return xmlout
    if schema is not None:
        if schema.startswith("FACTUR-X_"):
            schema = schema.replace("FACTUR-X_", "Factur-X_1.0.07_")
            schema = schema.replace("BASIC-WL", "BASICWL")
        schema = etree.XMLSchema(
            file=os.path.join(os.path.dirname(__file__), "schema", schema + ".xsd")
        )
    parser = etree.XMLParser(schema=schema)
    xml_root = etree.fromstring(xmlout, parser)

    return etree.tostring(
        xml_root, pretty_print=True, xml_declaration=True, encoding="UTF-8"
    )



from . import COMFORT, EXTENDED, NS_RAM
from .container import Container
from .elements import Element
from .fields import (
    ClassificationField,
    Field,
    IDField,
    MultiField,
    QuantityField,
    StringField,
)


class ProductCharacteristic(Element):
    type_code = StringField(
        NS_RAM,
        "TypeCode",
        required=False,
        profile=EXTENDED,
        _d="Art der Produkteigenschaft",
    )
    description = StringField(NS_RAM, "Description", required=True, profile=COMFORT)
    value_measure = QuantityField(
        NS_RAM,
        "ValueMeasure",
        required=False,
        profile=EXTENDED,
        _d="Numerische Messgröße",
    )
    value = StringField(NS_RAM, "Value", required=False, profile=COMFORT)

    class Meta:
        namespace = NS_RAM
        tag = "ApplicableProductCharacteristic"


class ProductClassification(Element):
    class_code = ClassificationField(
        NS_RAM, "ClassCode", required=False, profile=COMFORT
    )
    value = StringField(NS_RAM, "ClassName", required=False, profile=EXTENDED)

    class Meta:
        namespace = NS_RAM
        tag = "DesignatedProductClassification"


class ProductInstance(Element):
    batch_id = IDField(NS_RAM, "BatchID", required=False, profile=EXTENDED)
    serial_id = StringField(
        NS_RAM, "SupplierAssignedSerialID", required=False, profile=EXTENDED
    )

    class Meta:
        namespace = NS_RAM
        tag = "IndividualTradeProductInstance"


class OriginCountry(Element):
    id = StringField(
        NS_RAM, "ID", required=True, profile=EXTENDED, _d="Land der Produktherkunft"
    )

    class Meta:
        namespace = NS_RAM
        tag = "OriginTradeCountry"


class ReferencedProduct(Element):
    global_id = IDField(NS_RAM, "GlobalID", required=False, profile=EXTENDED)
    seller_assigned_id = StringField(
        NS_RAM, "SellerAssignedID", required=True, profile=EXTENDED
    )
    buyer_assigned_id = StringField(
        NS_RAM, "BuyerAssignedID", required=True, profile=EXTENDED
    )
    name = StringField(NS_RAM, "Name", required=False, profile=EXTENDED)
    description = StringField(NS_RAM, "Description", required=False, profile=EXTENDED)
    unit_quantity = QuantityField(
        NS_RAM, "UnitQuantity", required=False, profile=EXTENDED
    )

    class Meta:
        namespace = NS_RAM
        tag = "IncludedReferencedProduct"


class TradeProduct(Element):
    id = IDField(NS_RAM, "ID", required=False, profile=EXTENDED)
    global_id = IDField(NS_RAM, "GlobalID", required=False)
    seller_assigned_id = IDField(
        NS_RAM, "SellerAssignedID", required=False, profile=COMFORT
    )
    buyer_assigned_id = IDField(
        NS_RAM, "BuyerAssignedID", required=False, profile=COMFORT
    )
    industry_assigned_id = IDField(
        NS_RAM, "IndustryAssignedID", required=False, profile=EXTENDED
    )
    model_id = IDField(NS_RAM, "ModelID", required=False, profile=EXTENDED)
    name = StringField(NS_RAM, "Name", required=False)
    description = StringField(NS_RAM, "Description", required=False, profile=COMFORT)
    batch_id = IDField(NS_RAM, "BatchID", required=False, profile=EXTENDED)
    brand_name = StringField(NS_RAM, "BrandName", required=False, profile=EXTENDED)
    model_name = StringField(NS_RAM, "ModelName", required=False, profile=EXTENDED)
    characteristics: Container = MultiField(
        ProductCharacteristic, required=False, profile=COMFORT
    )
    classifications: Container = MultiField(
        ProductClassification, required=False, profile=COMFORT
    )
    instance: Container = MultiField(ProductInstance, required=False, profile=EXTENDED)
    origin: OriginCountry = Field(OriginCountry, required=False, profile=COMFORT)
    included_products: Container = MultiField(
        ReferencedProduct, required=False, profile=EXTENDED
    )

    class Meta:
        namespace = NS_RAM
        tag = "SpecifiedTradeProduct"



from codecs import open
from os import path
from setuptools import find_packages, setup

from drafthorse import version

here = path.abspath(path.dirname(__file__))

try:
    # Get the long description from the relevant file
    with open(path.join(here, "README.rst"), encoding="utf-8") as f:
        long_description = f.read()
except:  # noqa
    long_description = ""

setup(
    name="drafthorse",
    version=version,
    description="Python ZUGFeRD XML implementation",
    long_description=long_description,
    url="https://github.com/pretix/python-drafthorse",
    author="Raphael Michel",
    author_email="michel@rami.io",
    license="Apache License",
    classifiers=[
        "Intended Audience :: Developers",
        "Intended Audience :: Other Audience",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    keywords="xml banking sepa",
    install_requires=["lxml", "pypdf"],
    packages=find_packages(include=["drafthorse", "drafthorse.*"]),
    include_package_data=True,
)



from . import BASIC, COMFORT, EXTENDED, NS_RAM
from .elements import Element
from .container import StringContainer, CurrencyContainer
from .fields import (
    DateTimeField,
    DecimalField,
    Field,
    MultiCurrencyField,
    MultiStringField,
    QuantityField,
    StringField,
)


class FinancialCard(Element):
    id = StringField(NS_RAM, "ID")
    cardholder_name = StringField(NS_RAM, "CardholderName")

    class Meta:
        namespace = NS_RAM
        tag = "ApplicableTradeSettlementFinancialCard"


class PayerFinancialAccount(Element):
    iban = StringField(NS_RAM, "IBANID")

    class Meta:
        namespace = NS_RAM
        tag = "PayerPartyDebtorFinancialAccount"


class PayeeFinancialAccount(Element):
    iban = StringField(NS_RAM, "IBANID")
    account_name = StringField(NS_RAM, "AccountName")
    proprietary_id = StringField(NS_RAM, "ProprietaryID")

    class Meta:
        namespace = NS_RAM
        tag = "PayeePartyCreditorFinancialAccount"


class PayeeFinancialInstitution(Element):
    bic = StringField(NS_RAM, "BICID")

    class Meta:
        namespace = NS_RAM
        tag = "PayeeSpecifiedCreditorFinancialInstitution"


class PaymentMeans(Element):
    type_code = StringField(NS_RAM, "TypeCode", required=True, profile=COMFORT)
    information: StringContainer = MultiStringField(
        NS_RAM, "Information", required=False, profile=COMFORT
    )
    financial_card: FinancialCard = Field(FinancialCard)
    payer_account: PayerFinancialAccount = Field(PayerFinancialAccount)
    payee_account: PayeeFinancialAccount = Field(PayeeFinancialAccount)
    payee_institution: PayeeFinancialInstitution = Field(PayeeFinancialInstitution)

    class Meta:
        namespace = NS_RAM
        tag = "SpecifiedTradeSettlementPaymentMeans"


class PaymentPenaltyTerms(Element):
    basis_date_time = DateTimeField(
        NS_RAM,
        "BasisDateTime",
        required=False,
        profile=EXTENDED,
        _d="Bezugsdatum der Fälligkeit",
    )
    basis_period_measure = QuantityField(
        NS_RAM,
        "BasisPeriodMeasure",
        required=False,
        profile=EXTENDED,
        _d="Fälligkeitszeitraum",
    )
    basis_amount = DecimalField(
        NS_RAM,
        "BasisAmount",
        required=False,
        profile=EXTENDED,
        _d="Basisbetrag des Zahlungszuschlags",
    )
    calculation_percent = DecimalField(
        NS_RAM,
        "CalculationPercent",
        required=False,
        profile=EXTENDED,
        _d="Prozentwert des Zahlungszuschlags",
    )
    actual_amount = DecimalField(
        NS_RAM,
        "ActualPenaltyAmount",
        required=False,
        profile=EXTENDED,
        _d="Betrag des Zahlungszuschlags",
    )

    class Meta:
        namespace = NS_RAM
        tag = "ApplicableTradePaymentPenaltyTerms"


class PaymentDiscountTerms(Element):
    basis_date_time = DateTimeField(
        NS_RAM,
        "BasisDateTime",
        required=False,
        profile=EXTENDED,
        _d="Bezugsdatum der Fälligkeit",
    )
    basis_period_measure = QuantityField(
        NS_RAM,
        "BasisPeriodMeasure",
        required=False,
        profile=EXTENDED,
        _d="Fälligkeitszeitraum",
    )
    basis_amount = DecimalField(
        NS_RAM,
        "BasisAmount",
        required=False,
        profile=EXTENDED,
        _d="Basisbetrag des Zahlungsabschlags",
    )
    calculation_percent = DecimalField(
        NS_RAM,
        "CalculationPercent",
        required=False,
        profile=EXTENDED,
        _d="Prozentwert des Zahlungsabschlags",
    )
    actual_amount = DecimalField(
        NS_RAM,
        "ActualDiscountAmount",
        required=False,
        profile=EXTENDED,
        _d="Betrag des Zahlungsabschlags",
    )

    class Meta:
        namespace = NS_RAM
        tag = "ApplicableTradePaymentDiscountTerms"


class TaxApplicableTradeCurrencyExchange(Element):
    source_currency = StringField(
        NS_RAM, "SourceCurrencyCode", required=False, profile=EXTENDED
    )
    target_currency = StringField(
        NS_RAM, "TargetCurrencyCode", required=False, profile=EXTENDED
    )
    conversion_rate = DecimalField(
        NS_RAM, "ConversionRate", required=False, profile=EXTENDED
    )
    date_time_string = DateTimeField(
        NS_RAM, "ConversionRateDateTime", required=True, profile=EXTENDED
    )

    class Meta:
        namespace = NS_RAM
        tag = "TaxApplicableTradeCurrencyExchange"


class PaymentTerms(Element):
    description = StringField(
        NS_RAM,
        "Description",
        required=True,
        profile=COMFORT,
        _d="Freitext der Zahlungsbedingungen",
    )
    due = DateTimeField(
        NS_RAM,
        "DueDateDateTime",
        required=False,
        profile=COMFORT,
        _d="Fälligkeitsdatum",
    )
    debit_mandate_id = StringField(
        NS_RAM, "DirectDebitMandateID", required=False, profile=BASIC
    )
    partial_amount: CurrencyContainer = MultiCurrencyField(
        NS_RAM,
        "PartialPaymentAmount",
        profile=EXTENDED,
        required=False,
        _d="Betrag der Teilzahlung",
    )
    penalty_terms: PaymentPenaltyTerms = Field(
        PaymentPenaltyTerms,
        required=False,
        profile=EXTENDED,
        _d="Detailinformationen zu Zahlungszuschlägen",
    )
    discount_terms: PaymentDiscountTerms = Field(
        PaymentDiscountTerms,
        required=False,
        profile=EXTENDED,
        _d="Detailinformationen zu Zahlungsabschlägen",
    )

    class Meta:
        namespace = NS_RAM
        tag = "SpecifiedTradePaymentTerms"



from . import BASIC, COMFORT, EXTENDED, NS_RAM
from .elements import Element
from .container import Container, IDContainer
from .fields import Field, IDField, MultiField, MultiIDField, StringField


class PostalTradeAddress(Element):
    postcode = StringField(NS_RAM, "PostcodeCode", required=False, profile=BASIC)
    line_one = StringField(NS_RAM, "LineOne", required=False, profile=BASIC)
    line_two = StringField(NS_RAM, "LineTwo", required=False, profile=BASIC)
    line_three = StringField(NS_RAM, "LineThree", required=False, profile=BASIC)
    city_name = StringField(NS_RAM, "CityName", required=False, profile=BASIC)
    country_id = StringField(NS_RAM, "CountryID", required=True, profile=BASIC)
    country_subdivision = StringField(
        NS_RAM, "CountrySubDivisionName", required=False, profile=BASIC
    )

    class Meta:
        namespace = NS_RAM
        tag = "PostalTradeAddress"


class URIUniversalCommunication(Element):
    uri_ID = IDField(NS_RAM, "URIID", required=False, profile=BASIC)

    class Meta:
        namespace = NS_RAM
        tag = "URIUniversalCommunication"


class TaxRegistration(Element):
    id = IDField(NS_RAM, "ID")

    class Meta:
        namespace = NS_RAM
        tag = "SpecifiedTaxRegistration"


class PhoneNumber(Element):
    number = StringField(NS_RAM, "CompleteNumber", required=False, profile=EXTENDED)

    class Meta:
        namespace = NS_RAM
        tag = "TelephoneUniversalCommunication"


class FaxNumber(Element):
    number = StringField(NS_RAM, "CompleteNumber", required=False, profile=EXTENDED)

    class Meta:
        namespace = NS_RAM
        tag = "FaxUniversalCommunication"


class EmailURI(Element):
    address = StringField(NS_RAM, "URIID", required=False, profile=EXTENDED)

    class Meta:
        namespace = NS_RAM
        tag = "EmailURIUniversalCommunication"


class LegalOrganization(Element):
    id = IDField(NS_RAM, "ID", required=False, profile=BASIC)
    trade_name = StringField(
        NS_RAM,
        "TradingBusinessName",
        required=False,
        profile=BASIC,
        _d="Firmenname, sofern abweichend vom Namen",
    )
    trade_address: PostalTradeAddress = Field(
        PostalTradeAddress, required=False, profile=EXTENDED
    )

    class Meta:
        namespace = NS_RAM
        tag = "SpecifiedLegalOrganization"


class TradeContact(Element):
    person_name = StringField(NS_RAM, "PersonName", required=False, profile=EXTENDED)
    department_name = StringField(
        NS_RAM, "DepartmentName", required=False, profile=EXTENDED
    )
    telephone: PhoneNumber = Field(PhoneNumber, required=False, profile=EXTENDED)
    fax: FaxNumber = Field(FaxNumber, required=False, profile=EXTENDED)
    email: EmailURI = Field(EmailURI, required=False, profile=EXTENDED)

    class Meta:
        namespace = NS_RAM
        tag = "DefinedTradeContact"


class TradeParty(Element):
    id = StringField(
        NS_RAM, "ID", required=False, profile=COMFORT, _d="Kennung des Handelspartners"
    )
    global_id: IDContainer = MultiIDField(
        NS_RAM,
        "GlobalID",
        required=False,
        profile=COMFORT,
        _d="Globale Kennung des Handelspartners",
    )
    name = StringField(NS_RAM, "Name", required=False, profile=BASIC)
    role_code = StringField(NS_RAM, "RoleCode", required=False, profile=EXTENDED)
    description = StringField(
        NS_RAM,
        "Description",
        required=True,
        profile=COMFORT,
        _d="Zusätzliche rechliche Informationen des Handelspartners",
    )
    legal_organization: LegalOrganization = Field(
        LegalOrganization,
        required=False,
        profile=BASIC,
        _d="Handelsinformationen des Handelspartners",
    )
    contact: TradeContact = Field(
        TradeContact,
        required=False,
        profile=EXTENDED,
        _d="Ansprechpartner des Handelspartners",
    )
    address: PostalTradeAddress = Field(
        PostalTradeAddress,
        required=False,
        profile=BASIC,
        _d="Anschrift des Handelspartners",
    )
    electronic_address: Container = Field(
        URIUniversalCommunication, required=False, profile=BASIC
    )
    tax_registrations: Container = MultiField(
        TaxRegistration, required=False, profile=BASIC
    )


class SellerTaxRepresentativeTradeParty(TradeParty):
    class Meta:
        namespace = NS_RAM
        tag = "SellerTaxRepresentativeTradeParty"


class PayeeTradeParty(TradeParty):
    class Meta:
        namespace = NS_RAM
        tag = "PayeeTradeParty"


class PayerTradeParty(TradeParty):
    class Meta:
        namespace = NS_RAM
        tag = "PayerTradeParty"


class InvoicerTradeParty(TradeParty):
    class Meta:
        namespace = NS_RAM
        tag = "InvoicerTradeParty"


class InvoiceeTradeParty(TradeParty):
    class Meta:
        namespace = NS_RAM
        tag = "InvoiceeTradeParty"


class BuyerTradeParty(TradeParty):
    class Meta:
        namespace = NS_RAM
        tag = "BuyerTradeParty"


class SellerTradeParty(TradeParty):
    class Meta:
        namespace = NS_RAM
        tag = "SellerTradeParty"


class EndUserTradeParty(TradeParty):
    class Meta:
        namespace = NS_RAM
        tag = "ProductEndUserTradeParty"


class ShipToTradeParty(TradeParty):
    class Meta:
        namespace = NS_RAM
        tag = "ShipToTradeParty"


class ShipFromTradeParty(TradeParty):
    class Meta:
        namespace = NS_RAM
        tag = "ShipFromTradeParty"


class UltimateShipToTradeParty(TradeParty):
    class Meta:
        namespace = NS_RAM
        tag = "UltimateShipToTradeParty"



from decimal import Decimal

from . import BASIC, NS_UDT
from .container import (
    Container,
    CurrencyContainer,
    IDContainer,
    StringContainer,
)


class Field:
    def __init__(self, cls, default=False, required=False, profile=BASIC, _d=None):
        self.cls = cls
        self.required = required
        self.default = default
        self.__doc__ = _d
        self.profile = profile
        super().__init__()

    def initialize(self):
        return self.cls(required=self.required)

    def __get__(self, instance, objtype):
        if instance._data.get(self.name, None) is None:
            instance._data[self.name] = self.initialize()
        return instance._data[self.name]

    def __set__(self, instance, value):
        raise AttributeError("Read-only!")

    def __delete__(self, instance):
        del instance._data[self.name]

    def __set_name__(self, owner, name):
        self.name = name


class StringField(Field):
    def __init__(
        self, namespace, tag, default=False, required=False, profile=BASIC, _d=None
    ):
        from .elements import StringElement

        super().__init__(StringElement, default, required, profile, _d)
        self.namespace = namespace
        self.tag = tag

    def initialize(self):
        return self.cls(self.namespace, self.tag)

    def __set__(self, instance, value):
        if instance._data.get(self.name, None) is None:
            instance._data[self.name] = self.initialize()
        instance._data[self.name]._text = value


class AgencyIDField(Field):
    def __init__(
        self, namespace, tag, default=False, required=False, profile=BASIC, _d=None
    ):
        from .elements import AgencyIDElement

        super().__init__(AgencyIDElement, default, required, profile, _d)
        self.namespace = namespace
        self.tag = tag

    def initialize(self):
        return self.cls(self.namespace, self.tag)

    def __set__(self, instance, value):
        if instance._data.get(self.name, None) is None:
            instance._data[self.name] = self.initialize()

        if not isinstance(value, (tuple, list)):
            raise TypeError("Please pass a 2-tuple of scheme agency ID and ID.")
        instance._data[self.name]._text = value[1]
        instance._data[self.name]._scheme_id = value[0]


class ClassificationField(Field):
    def __init__(
        self, namespace, tag, default=False, required=False, profile=BASIC, _d=None
    ):
        from .elements import ClassificationElement

        super().__init__(ClassificationElement, default, required, profile, _d)
        self.namespace = namespace
        self.tag = tag

    def initialize(self):
        return self.cls(self.namespace, self.tag)

    def __set__(self, instance, value):
        if instance._data.get(self.name, None) is None:
            instance._data[self.name] = self.initialize()

        if not isinstance(value, (tuple, list)):
            raise TypeError("Please pass a 3-tuple of list ID, list version and ID.")
        instance._data[self.name]._text = value[2]
        instance._data[self.name]._list_id = value[0]
        instance._data[self.name]._list_version_id = value[1]


class IDField(Field):
    def __init__(
        self, namespace, tag, default=False, required=False, profile=BASIC, _d=None
    ):
        from .elements import IDElement

        super().__init__(IDElement, default, required, profile, _d)
        self.namespace = namespace
        self.tag = tag

    def initialize(self):
        return self.cls(self.namespace, self.tag)

    def __set__(self, instance, value):
        if instance._data.get(self.name, None) is None:
            instance._data[self.name] = self.initialize()

        if isinstance(value, (tuple, list)):
            if len(value) == 2:
                instance._data[self.name]._text = value[1]
                instance._data[self.name]._scheme_id = value[0]
            else:
                raise TypeError(
                    "Please pass a 2-tuple of including scheme ID and ID, or just an ID."
                )
        else:
            instance._data[self.name]._text = value


class CurrencyField(Field):
    def __init__(
        self, namespace, tag, default=False, required=False, profile=BASIC, _d=None
    ):
        from .elements import CurrencyElement

        super().__init__(CurrencyElement, default, required, profile, _d)
        self.namespace = namespace
        self.tag = tag

    def __set__(self, instance, value):
        if instance._data.get(self.name, None) is None:
            instance._data[self.name] = self.initialize()

        if isinstance(value, (int, float, Decimal)):
            value = (value, None)
        if not isinstance(value, (tuple, list)):
            raise TypeError("Please pass a 2-tuple of including amount and currency.")

        instance._data[self.name]._amount = value[0]
        instance._data[self.name]._currency = value[1]

    def initialize(self):
        return self.cls(self.namespace, self.tag)


class DecimalField(Field):
    def __init__(
        self, namespace, tag, default=False, required=False, profile=BASIC, _d=None
    ):
        from .elements import DecimalElement

        super().__init__(DecimalElement, default, required, profile, _d)
        self.namespace = namespace
        self.tag = tag

    def __set__(self, instance, value):
        if instance._data.get(self.name, None) is None:
            instance._data[self.name] = self.initialize()
        instance._data[self.name]._value = value

    def initialize(self):
        return self.cls(self.namespace, self.tag)


class QuantityField(Field):
    def __init__(
        self, namespace, tag, default=False, required=False, profile=BASIC, _d=None
    ):
        from .elements import QuantityElement

        super().__init__(QuantityElement, default, required, profile, _d)
        self.namespace = namespace
        self.tag = tag

    def __set__(self, instance, value):
        if instance._data.get(self.name, None) is None:
            instance._data[self.name] = self.initialize()

        if not isinstance(value, (tuple, list)):
            raise TypeError("Please pass a 2-tuple of including amount and unit code.")
        instance._data[self.name]._amount = value[0]
        instance._data[self.name]._unit_code = value[1]

    def initialize(self):
        return self.cls(self.namespace, self.tag)


class BinaryObjectField(Field):
    def __init__(
        self, namespace, tag, default=False, required=False, profile=BASIC, _d=None
    ):
        from .elements import BinaryObjectElement

        super().__init__(BinaryObjectElement, default, required, profile, _d)
        self.namespace = namespace
        self.tag = tag

    def __set__(self, instance, value):
        if instance._data.get(self.name, None) is None:
            instance._data[self.name] = self.initialize()

        if not isinstance(value, (tuple, list)):
            raise TypeError(
                "Please pass a 3-tuple of mimeCode, filename and base64-encoded binary."
            )
        instance._data[self.name]._text = value[2]
        instance._data[self.name]._filename = value[1]
        instance._data[self.name]._mime_code = value[0]

    def initialize(self):
        return self.cls(self.namespace, self.tag)


class IndicatorField(Field):
    def __init__(
        self, namespace, tag, default=False, required=False, profile=BASIC, _d=None
    ):
        from .elements import IndicatorElement

        super().__init__(IndicatorElement, default, required, profile, _d)
        self.namespace = namespace
        self.tag = tag

    def __set__(self, instance, value):
        if instance._data.get(self.name, None) is None:
            instance._data[self.name] = self.initialize()
        instance._data[self.name]._value = value

    def initialize(self):
        return self.cls(self.namespace, self.tag)


class DateTimeField(Field):
    def __init__(
        self,
        namespace,
        tag,
        default=False,
        required=False,
        profile=BASIC,
        _d=None,
        date_time_namespace=NS_UDT,
    ):
        from .elements import DateTimeElement

        super().__init__(DateTimeElement, default, required, profile, _d)
        self.namespace = namespace
        self.tag = tag
        self._date_time_namespace = date_time_namespace

    def __set__(self, instance, value):
        if instance._data.get(self.name, None) is None:
            instance._data[self.name] = self.initialize()
        instance._data[self.name]._value = value

    def initialize(self):
        return self.cls(
            self.namespace, self.tag, date_time_namespace=self._date_time_namespace
        )


class DirectDateTimeField(Field):
    def __init__(
        self, namespace, tag, default=False, required=False, profile=BASIC, _d=None
    ):
        from .elements import DirectDateTimeElement

        super().__init__(DirectDateTimeElement, default, required, profile, _d)
        self.namespace = namespace
        self.tag = tag

    def __set__(self, instance, value):
        if instance._data.get(self.name, None) is None:
            instance._data[self.name] = self.initialize()
        instance._data[self.name]._value = value

    def initialize(self):
        return self.cls(self.namespace, self.tag)


class MultiField(Field):
    def __init__(
        self, inner_type, default=False, required=False, profile=BASIC, _d=None
    ):
        super().__init__(Container, default, required, profile, _d)
        self.inner_type = inner_type

    def initialize(self):
        return self.cls(child_type=self.inner_type)


class MultiStringField(Field):
    def __init__(
        self, namespace, tag, default=False, required=False, profile=BASIC, _d=None
    ):
        super().__init__(StringContainer, default, required, profile, _d)
        self.namespace = namespace
        self.tag = tag

    def initialize(self):
        return self.cls(child_type=str, namespace=self.namespace, tag=self.tag)


class MultiCurrencyField(Field):
    def __init__(
        self, namespace, tag, default=False, required=False, profile=BASIC, _d=None
    ):
        super().__init__(CurrencyContainer, default, required, profile, _d)
        self.namespace = namespace
        self.tag = tag

    def initialize(self):
        return self.cls(
            child_type=(tuple, list), namespace=self.namespace, tag=self.tag
        )


class MultiIDField(Field):
    def __init__(
        self, namespace, tag, default=False, required=False, profile=BASIC, _d=None
    ):
        super().__init__(IDContainer, default, required, profile, _d)
        self.namespace = namespace
        self.tag = tag

    def initialize(self):
        return self.cls(
            child_type=(tuple, list), namespace=self.namespace, tag=self.tag
        )
